/**
 * @file libgalago-gtk/galago-gtk-service-option-menu.c
 *       Service OptionMenu widget
 *
 * @Copyright (C) 2004-2005 Christian Hammond.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#include <libgalago-gtk/galago-gtk-service-option-menu.h>
#include <libgalago-gtk/galago-gtk-service-menu-item.h>
#include <libgalago-gtk/galago-gdk-pixbuf.h>
#include <libgalago/galago.h>
#include <string.h>

struct _GalagoGtkServiceOptionMenuPriv
{
	unsigned long service_added_id;
	unsigned long service_removed_id;
};

enum
{
	SERVICE_CHANGED,
	LAST_SIGNAL
};

enum
{
	COLUMN_DATA,
	COLUMN_ICON,
	COLUMN_NAME,
	NUM_COLUMNS
};

static void galago_gtk_service_option_menu_class_init(
	GalagoGtkServiceOptionMenuClass *klass);
static void galago_gtk_service_option_menu_init(
	GalagoGtkServiceOptionMenu *optmenu);
static void galago_gtk_service_option_menu_finalize(GObject *obj);
static void galago_gtk_service_option_menu_destroy(GtkObject *obj);
static void service_added_cb(void *obj, const char *property,
							 GalagoService *service,
							 GalagoGtkServiceOptionMenu *optmenu);
static void service_removed_cb(void *obj, const char *property,
							   GalagoService *service,
							   GalagoGtkServiceOptionMenu *optmenu);

static GtkOptionMenuClass *parent_class = NULL;
static guint signals[LAST_SIGNAL] = {0};

GType
galago_gtk_service_option_menu_get_type(void)
{
	static GType type = 0;

	if (!type)
	{
		static const GTypeInfo info =
		{
			sizeof(GalagoGtkServiceOptionMenuClass),
			NULL,
			NULL,
			(GClassInitFunc)galago_gtk_service_option_menu_class_init,
			NULL,
			NULL,
			sizeof(GalagoGtkServiceOptionMenu),
			0,
			(GInstanceInitFunc)galago_gtk_service_option_menu_init
		};

		type = g_type_register_static(GTK_TYPE_OPTION_MENU,
									  "GalagoGtkServiceOptionMenu", &info, 0);
	}

	return type;
}

static void
galago_gtk_service_option_menu_class_init(
	GalagoGtkServiceOptionMenuClass *klass)
{
	GObjectClass *gobject_class;
	GtkObjectClass *object_class;

	parent_class = g_type_class_peek_parent(klass);

	gobject_class = G_OBJECT_CLASS(klass);
	object_class  = GTK_OBJECT_CLASS(klass);

	gobject_class->finalize = galago_gtk_service_option_menu_finalize;

	object_class->destroy = galago_gtk_service_option_menu_destroy;

	signals[SERVICE_CHANGED] =
		g_signal_new("service_changed",
					 G_TYPE_FROM_CLASS(gobject_class),
					 G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
					 G_STRUCT_OFFSET(GalagoGtkServiceOptionMenuClass,
									 service_changed),
					 NULL, NULL,
					 gtk_marshal_VOID__POINTER,
					 G_TYPE_NONE, 1,
					 G_TYPE_POINTER);
}

static void
option_menu_changed(GalagoGtkServiceOptionMenu *optmenu, gpointer user_data)
{
	GalagoService *service;

	service = galago_gtk_service_option_menu_get_active(optmenu);

	g_signal_emit(optmenu, signals[SERVICE_CHANGED], 0, service);
}

static void
galago_gtk_service_option_menu_init(GalagoGtkServiceOptionMenu *optmenu)
{
	optmenu->priv = g_new0(GalagoGtkServiceOptionMenuPriv, 1);

	g_signal_connect(G_OBJECT(optmenu), "changed",
					 G_CALLBACK(option_menu_changed), NULL);

	optmenu->priv->service_added_id =
		galago_signal_connect(galago_get_core(), "service-added",
							  GALAGO_CALLBACK(service_added_cb),
							  optmenu);
	optmenu->priv->service_removed_id =
		galago_signal_connect(galago_get_core(), "service-removed",
							  GALAGO_CALLBACK(service_removed_cb),
							  optmenu);
}

static void
galago_gtk_service_option_menu_finalize(GObject *obj)
{
	GalagoGtkServiceOptionMenu *optmenu;

	g_return_if_fail(obj != NULL);
	g_return_if_fail(GALAGO_GTK_IS_SERVICE_OPTION_MENU(obj));

	optmenu = GALAGO_GTK_SERVICE_OPTION_MENU(obj);

	g_free(optmenu->priv);

	if (G_OBJECT_CLASS(parent_class)->finalize)
		G_OBJECT_CLASS(parent_class)->finalize(obj);
}

static void
galago_gtk_service_option_menu_destroy(GtkObject *obj)
{
	GalagoGtkServiceOptionMenu *optmenu;

	g_return_if_fail(obj != NULL);
	g_return_if_fail(GALAGO_GTK_IS_SERVICE_OPTION_MENU(obj));

	optmenu = GALAGO_GTK_SERVICE_OPTION_MENU(obj);

	if (optmenu->priv->service_added_id > 0)
		galago_signals_disconnect_by_id(optmenu->priv->service_added_id);

	if (optmenu->priv->service_removed_id > 0)
		galago_signals_disconnect_by_id(optmenu->priv->service_removed_id);

	if (GTK_OBJECT_CLASS(parent_class)->destroy)
		GTK_OBJECT_CLASS(parent_class)->destroy(obj);
}

static void
rebuild_list(GalagoGtkServiceOptionMenu *optmenu)
{
	const GalagoList *l;
	GtkWidget *menuitem;
	GtkWidget *menu;

	menu = gtk_menu_new();
	gtk_widget_show(menu);
	gtk_option_menu_set_menu(GTK_OPTION_MENU(optmenu), menu);

	for (l = galago_core_get_services(FALSE, TRUE); l != NULL; l = l->next)
	{
		GalagoService *service = (GalagoService *)l->data;

		menuitem = galago_gtk_service_menu_item_new(service);
		gtk_widget_show(menuitem);
		gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);
	}
}

static void
service_added_cb(void *obj, const char *property, GalagoService *service,
				 GalagoGtkServiceOptionMenu *optmenu)
{
	GalagoService *cur_service;

	cur_service = galago_gtk_service_option_menu_get_active(optmenu);

	rebuild_list(optmenu);

	if (cur_service != NULL)
	{
		if (galago_gtk_service_option_menu_get_active(optmenu) != cur_service)
			galago_gtk_service_option_menu_set_active(optmenu, cur_service);
	}
	else
		gtk_option_menu_set_history(GTK_OPTION_MENU(optmenu), 0);
}

static void
service_removed_cb(void *obj, const char *property, GalagoService *service,
				   GalagoGtkServiceOptionMenu *optmenu)
{
	GalagoService *cur_service;

	cur_service = galago_gtk_service_option_menu_get_active(optmenu);

	rebuild_list(optmenu);

	if (cur_service == service)
	{
		if (galago_gtk_service_option_menu_get_active(optmenu) != cur_service)
			galago_gtk_service_option_menu_set_active(optmenu, cur_service);
	}
	else
		gtk_option_menu_set_history(GTK_OPTION_MENU(optmenu), 0);
}

GtkWidget *
galago_gtk_service_option_menu_new(void)
{
	GalagoGtkServiceOptionMenu *optmenu;

	optmenu = g_object_new(GALAGO_GTK_TYPE_SERVICE_OPTION_MENU, NULL);

	rebuild_list(optmenu);

	gtk_option_menu_set_history(GTK_OPTION_MENU(optmenu), 0);

	return GTK_WIDGET(optmenu);
}

void
galago_gtk_service_option_menu_set_active(GalagoGtkServiceOptionMenu *optmenu,
										  const GalagoService *service)
{
	GtkWidget *menu;
	GList *l;
	int i;

	g_return_if_fail(optmenu != NULL);
	g_return_if_fail(GALAGO_GTK_IS_SERVICE_OPTION_MENU(optmenu));
	g_return_if_fail(service != NULL);

	menu = gtk_option_menu_get_menu(GTK_OPTION_MENU(optmenu));

	for (l = gtk_container_get_children(GTK_CONTAINER(menu)), i = 0;
		 l != NULL;
		 l = l->next, i++)
	{
		GtkWidget *menuitem = GTK_WIDGET(l);
		GalagoService *temp_service;

		if (!GALAGO_GTK_IS_SERVICE_MENU_ITEM(menuitem))
			continue;

		temp_service = galago_gtk_service_menu_item_get_service(
			GALAGO_GTK_SERVICE_MENU_ITEM(menuitem));

		if (temp_service == service)
		{
			gtk_option_menu_set_history(GTK_OPTION_MENU(optmenu), i);
			break;
		}
	}
}

GalagoService *
galago_gtk_service_option_menu_get_active(
	const GalagoGtkServiceOptionMenu *optmenu)
{
	GtkWidget *menu;
	GtkWidget *menuitem;

	g_return_val_if_fail(optmenu != NULL,                            NULL);
	g_return_val_if_fail(GALAGO_GTK_IS_SERVICE_OPTION_MENU(optmenu), NULL);

	menu = gtk_option_menu_get_menu(GTK_OPTION_MENU(optmenu));

	if (menu == NULL)
		return NULL;

	menuitem = gtk_menu_get_active(GTK_MENU(menu));

	g_return_val_if_fail(GALAGO_GTK_IS_SERVICE_MENU_ITEM(menuitem), NULL);

	return galago_gtk_service_menu_item_get_service(
		GALAGO_GTK_SERVICE_MENU_ITEM(menuitem));
}
