/*GTK - The GIMP Toolkit
 *Copyright(C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 *This library is free software; you can redistribute it and/or
 *modify it under the terms of the GNU Lesser General Public
 *License as published by the Free Software Foundation; either
 *version 2 of the License, or(at your option) any later version.
 *
 *This library is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *Lesser General Public License for more details.
 *
 *You should have received a copy of the GNU Lesser General Public
 *License along with this library; if not, write to the
 *Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *Boston, MA 02111-1307, USA.
 */

/*
 *Modified by the GTK+ Team and others 1997-2000.  See the AUTHORS
 *file for a list of people on the GTK+ Team.  See the ChangeLog
 *files for a list of changes.  These files are distributed with
 *GTK+ at ftp://ftp.gtk.org/pub/gtk/.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <gtk/gtklabel.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkmenuitem.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtkstyle.h>

#include "galago-gtk-tooltips.h"


#define DEFAULT_DELAY 500	/*Default delay in ms */
#define STICKY_DELAY 0		/*Delay before popping up next tip
				 *if we are sticky
				 */
#define STICKY_REVERT_DELAY 1000	/*Delay before sticky tooltips revert
					 *to normal
					 */

static void galago_gtk_tooltips_class_init(GalagoGtkTooltipsClass *klass);
static void galago_gtk_tooltips_init(GalagoGtkTooltips *tooltips);
static void galago_gtk_tooltips_destroy(GtkObject *object);

static void galago_gtk_tooltips_event_handler(GtkWidget *widget,
					  GdkEvent *event);
static void galago_gtk_tooltips_widget_unmap(GtkWidget *widget,
					 gpointer data);
static void galago_gtk_tooltips_widget_remove(GtkWidget *widget,
					  gpointer data);
static void galago_gtk_tooltips_set_active_widget(GalagoGtkTooltips *tooltips,
					      GtkWidget *widget);
static gint galago_gtk_tooltips_timeout(gpointer data);

static gint galago_gtk_tooltips_paint_window(GalagoGtkTooltips *tooltips);
static void galago_gtk_tooltips_draw_tips(GalagoGtkTooltips *tooltips);
static void galago_gtk_tooltips_unset_tip_window(GalagoGtkTooltips *tooltips);

static gboolean get_keyboard_mode(GtkWidget *widget);

static GtkObjectClass *parent_class;
static const gchar *tooltips_data_key = "_GalagoGtkTooltipsData";

GType
galago_gtk_tooltips_get_type(void)
{
	static GType tooltips_type = 0;

	if (!tooltips_type)
	{
		static const GTypeInfo tooltips_info =
		{
			sizeof(GalagoGtkTooltipsClass),
			NULL,
			NULL,
			(GClassInitFunc)galago_gtk_tooltips_class_init,
			NULL,
			NULL,
			sizeof(GalagoGtkTooltips),
			0,
			(GInstanceInitFunc)galago_gtk_tooltips_init,
		};

		tooltips_type = g_type_register_static(GTK_TYPE_OBJECT,
											   "GalagoGtkTooltips",
											   &tooltips_info, 0);
	}

	return tooltips_type;
}

static void
galago_gtk_tooltips_class_init(GalagoGtkTooltipsClass *class)
{
	GtkObjectClass *object_class;

	object_class = (GtkObjectClass *)class;

	parent_class = g_type_class_peek_parent(class);

	object_class->destroy = galago_gtk_tooltips_destroy;
}

static void
galago_gtk_tooltips_init(GalagoGtkTooltips *tooltips)
{
	tooltips->tip_window = NULL;
	tooltips->active_tips_data = NULL;
	tooltips->tips_data_list = NULL;

	tooltips->delay = DEFAULT_DELAY;
	tooltips->enabled = TRUE;
	tooltips->timer_tag = 0;
	tooltips->use_sticky_delay = FALSE;
	tooltips->last_popdown.tv_sec = -1;
	tooltips->last_popdown.tv_usec = -1;
}

GalagoGtkTooltips *
galago_gtk_tooltips_new(void)
{
	return g_object_new(GALAGO_GTK_TYPE_TOOLTIPS, NULL);
}

static void
galago_gtk_tooltips_destroy_data(GalagoGtkTooltipsData *tooltips_data)
{
	if (tooltips_data->tip_text != NULL)
		g_free(tooltips_data->tip_text);

	if (tooltips_data->tip_private != NULL)
		g_free(tooltips_data->tip_private);

	g_signal_handlers_disconnect_by_func(tooltips_data->widget,
										 galago_gtk_tooltips_event_handler,
										 tooltips_data);
	g_signal_handlers_disconnect_by_func(tooltips_data->widget,
										 galago_gtk_tooltips_widget_unmap,
										 tooltips_data);
	g_signal_handlers_disconnect_by_func(tooltips_data->widget,
										 galago_gtk_tooltips_widget_remove,
										 tooltips_data);

	g_object_set_data(G_OBJECT(tooltips_data->widget), tooltips_data_key, NULL);
	g_object_unref(tooltips_data->widget);
	g_free(tooltips_data);
}

static void
tip_window_display_closed(GdkDisplay *display, gboolean was_error,
						  GalagoGtkTooltips *tooltips)
{
	galago_gtk_tooltips_unset_tip_window(tooltips);
}

static void
disconnect_tip_window_display_closed(GalagoGtkTooltips *tooltips)
{
	g_signal_handlers_disconnect_by_func(
		gtk_widget_get_display(tooltips->tip_window),
		(gpointer)tip_window_display_closed, tooltips);
}

static void
galago_gtk_tooltips_unset_tip_window(GalagoGtkTooltips *tooltips)
{
	if (tooltips->tip_window)
	{
		disconnect_tip_window_display_closed(tooltips);

		gtk_widget_destroy(tooltips->tip_window);
		tooltips->tip_window = NULL;
	}
}

static void
galago_gtk_tooltips_destroy(GtkObject *object)
{
	GalagoGtkTooltips *tooltips = GALAGO_GTK_TOOLTIPS(object);
	GList *current;
	GalagoGtkTooltipsData *tooltips_data;

	g_return_if_fail(tooltips != NULL);

	if (tooltips->timer_tag)
	{
		g_source_remove(tooltips->timer_tag);
		tooltips->timer_tag = 0;
	}

	if (tooltips->tips_data_list != NULL)
	{
		for (current = g_list_first(tooltips->tips_data_list);
			 current != NULL;
			 current = current->next)
		{
			tooltips_data = (GalagoGtkTooltipsData *)current->data;
			galago_gtk_tooltips_widget_remove(tooltips_data->widget,
											  tooltips_data);
		}
	}

	galago_gtk_tooltips_unset_tip_window(tooltips);
}

static void
galago_gtk_tooltips_update_screen(GalagoGtkTooltips *tooltips,
								  gboolean new_window)
{
	gboolean screen_changed = FALSE;

	if (tooltips->active_tips_data &&
	    tooltips->active_tips_data->widget)
	{
		GdkScreen *screen =
		    gtk_widget_get_screen(tooltips->active_tips_data->widget);

		screen_changed =
			(screen != gtk_widget_get_screen(tooltips->tip_window));

		if (screen_changed)
		{
			if (!new_window)
				disconnect_tip_window_display_closed(tooltips);

			gtk_window_set_screen(GTK_WINDOW(tooltips->tip_window), screen);
		}
	}

	if (screen_changed || new_window)
	{
		g_signal_connect(gtk_widget_get_display(tooltips->tip_window),
						 "closed",
						 G_CALLBACK(tip_window_display_closed), tooltips);
	}
}

void
galago_gtk_tooltips_force_window(GalagoGtkTooltips *tooltips)
{
	g_return_if_fail(GALAGO_GTK_IS_TOOLTIPS(tooltips));

	if (!tooltips->tip_window)
	{
		tooltips->tip_window = gtk_window_new(GTK_WINDOW_POPUP);
		galago_gtk_tooltips_update_screen(tooltips, TRUE);
		gtk_widget_set_app_paintable(tooltips->tip_window, TRUE);
		gtk_window_set_resizable(GTK_WINDOW(tooltips->tip_window), FALSE);
		gtk_widget_set_name(tooltips->tip_window, "gtk-tooltips");
		gtk_container_set_border_width(GTK_CONTAINER(tooltips->tip_window), 4);

		g_signal_connect_swapped(tooltips->tip_window, "expose_event",
								 G_CALLBACK(galago_gtk_tooltips_paint_window),
								 tooltips);

		tooltips->tip_label = gtk_label_new(NULL);
		gtk_label_set_line_wrap(GTK_LABEL(tooltips->tip_label), TRUE);
		gtk_misc_set_alignment(GTK_MISC(tooltips->tip_label), 0.5, 0.5);
		gtk_widget_show(tooltips->tip_label);

		gtk_container_add(GTK_CONTAINER(tooltips->tip_window),
						  tooltips->tip_label);

		g_signal_connect(tooltips->tip_window, "destroy",
						 G_CALLBACK(gtk_widget_destroyed),
						 &tooltips->tip_window);
	}
}

void
galago_gtk_tooltips_enable(GalagoGtkTooltips *tooltips)
{
	g_return_if_fail(tooltips != NULL);

	tooltips->enabled = TRUE;
}

void
galago_gtk_tooltips_disable(GalagoGtkTooltips *tooltips)
{
	g_return_if_fail(tooltips != NULL);

	galago_gtk_tooltips_set_active_widget(tooltips, NULL);

	tooltips->enabled = FALSE;
}

GalagoGtkTooltipsData *
galago_gtk_tooltips_data_get(GtkWidget *widget)
{
	g_return_val_if_fail(widget != NULL, NULL);

	return g_object_get_data(G_OBJECT(widget), tooltips_data_key);
}

void
galago_gtk_tooltips_set_tip(GalagoGtkTooltips *tooltips, GtkWidget *widget,
							const gchar *tip_text, const gchar *tip_private)
{
	GalagoGtkTooltipsData *tooltips_data;

	g_return_if_fail(GALAGO_GTK_IS_TOOLTIPS(tooltips));
	g_return_if_fail(widget != NULL);

	tooltips_data = galago_gtk_tooltips_data_get(widget);

	if (!tip_text)
	{
		if (tooltips_data)
		{
			galago_gtk_tooltips_widget_remove(tooltips_data->widget,
											  tooltips_data);
		}

		return;
	}

	if (tooltips->active_tips_data
	    && tooltips->active_tips_data->widget == widget
	    && GTK_WIDGET_DRAWABLE(tooltips->active_tips_data->widget))
	{
		g_free(tooltips_data->tip_text);
		g_free(tooltips_data->tip_private);

		tooltips_data->tip_text = g_strdup(tip_text);
		tooltips_data->tip_private = g_strdup(tip_private);

		galago_gtk_tooltips_draw_tips(tooltips);
	}
	else
	{
		g_object_ref(widget);

		if (tooltips_data)
		{
			galago_gtk_tooltips_widget_remove(tooltips_data->widget,
											  tooltips_data);
		}

		tooltips_data = g_new0(GalagoGtkTooltipsData, 1);

		tooltips_data->tooltips = tooltips;
		tooltips_data->widget = widget;

		tooltips_data->tip_text = g_strdup(tip_text);
		tooltips_data->tip_private = g_strdup(tip_private);

		tooltips->tips_data_list = g_list_append(tooltips->tips_data_list,
												 tooltips_data);

		g_signal_connect_after(widget, "event-after",
							   G_CALLBACK(galago_gtk_tooltips_event_handler),
							   tooltips_data);

		g_object_set_data(G_OBJECT(widget), tooltips_data_key, tooltips_data);

		g_signal_connect(widget, "unmap",
						 G_CALLBACK(galago_gtk_tooltips_widget_unmap),
						 tooltips_data);

		g_signal_connect(widget, "unrealize",
						 G_CALLBACK(galago_gtk_tooltips_widget_unmap),
						 tooltips_data);

		g_signal_connect(widget, "destroy",
						 G_CALLBACK(galago_gtk_tooltips_widget_remove),
						 tooltips_data);
	}
}

void
galago_gtk_tooltips_set_tip_func(GalagoGtkTooltips *tooltips,
								 GtkWidget *widget,
								 GalagoGtkTooltipFunc func,
								 GtkWidget *func_widget)
{
	GalagoGtkTooltipsData *tooltips_data;

	g_return_if_fail(tooltips != NULL);
	g_return_if_fail(widget   != NULL);
	g_return_if_fail(GALAGO_GTK_IS_TOOLTIPS(tooltips));
	g_return_if_fail(GTK_IS_WIDGET(widget));

	if (func_widget == NULL)
		func_widget = widget;

	tooltips_data = galago_gtk_tooltips_data_get(widget);

	if (func == NULL)
	{
		if (tooltips_data == NULL)
		{
			galago_gtk_tooltips_widget_remove(tooltips_data->widget,
											  tooltips_data);
		}

		return;
	}

	if (tooltips->active_tips_data != NULL &&
		tooltips->active_tips_data->widget == widget &&
		GTK_WIDGET_DRAWABLE(tooltips->active_tips_data->widget))
	{
		if (tooltips_data->tip_text != NULL)
			g_free(tooltips_data->tip_text);

		if (tooltips_data->tip_private != NULL)
			g_free(tooltips_data->tip_private);

		tooltips_data->tip_func        = func;
		tooltips_data->tip_func_widget = func_widget;

		galago_gtk_tooltips_draw_tips(tooltips);
	}
	else
	{
		g_object_ref(widget);

		if (tooltips_data != NULL)
		{
			galago_gtk_tooltips_widget_remove(tooltips_data->widget,
											  tooltips_data);
		}

		tooltips_data = g_new0(GalagoGtkTooltipsData, 1);

		tooltips_data->tooltips = tooltips;
		tooltips_data->widget   = widget;

		tooltips_data->tip_func        = func;
		tooltips_data->tip_func_widget = func_widget;

		tooltips->tips_data_list = g_list_append(tooltips->tips_data_list,
												 tooltips_data);

		g_signal_connect_after(widget, "event-after",
							   G_CALLBACK(galago_gtk_tooltips_event_handler),
							   tooltips_data);

		g_object_set_data(G_OBJECT(widget), tooltips_data_key, tooltips_data);

		g_signal_connect(widget, "unmap",
						 G_CALLBACK(galago_gtk_tooltips_widget_unmap),
						 tooltips_data);

		g_signal_connect(widget, "unrealize",
						 G_CALLBACK(galago_gtk_tooltips_widget_unmap),
						 tooltips_data);

		g_signal_connect(widget, "destroy",
						 G_CALLBACK(galago_gtk_tooltips_widget_remove),
						 tooltips_data);

	}
}

static gint
galago_gtk_tooltips_paint_window(GalagoGtkTooltips *tooltips)
{
	gtk_paint_flat_box(tooltips->tip_window->style,
					   tooltips->tip_window->window, GTK_STATE_NORMAL,
					   GTK_SHADOW_OUT, NULL,
					   GTK_WIDGET(tooltips->tip_window), "tooltip",
					   0, 0, -1, -1);

	return FALSE;
}

static void
galago_gtk_tooltips_draw_tips(GalagoGtkTooltips *tooltips)
{
	GtkRequisition requisition;
	GtkWidget *widget;
	GtkStyle *style;
	gint x, y, w, h, scr_w, scr_h;
	GalagoGtkTooltipsData *data;
	gboolean keyboard_mode;
	GdkScreen *screen;

	if (!tooltips->tip_window)
		galago_gtk_tooltips_force_window(tooltips);
	else if (GTK_WIDGET_VISIBLE(tooltips->tip_window))
		g_get_current_time(&tooltips->last_popdown);

	gtk_widget_ensure_style(tooltips->tip_window);
	style = tooltips->tip_window->style;

	if (tooltips->active_tips_data->tip_func_widget != NULL)
		widget = tooltips->active_tips_data->tip_func_widget;
	else
		widget = tooltips->active_tips_data->widget;

	keyboard_mode = get_keyboard_mode(widget);

	galago_gtk_tooltips_update_screen(tooltips, FALSE);

	screen = gtk_widget_get_screen(widget);
	scr_w = gdk_screen_get_width(screen);
	scr_h = gdk_screen_get_height(screen);

	data = tooltips->active_tips_data;

	if (data->tip_func != NULL)
	{
		char *tip_text = data->tip_func(data->tip_func_widget);

		if (tip_text == NULL)
			return;

		gtk_label_set_markup(GTK_LABEL(tooltips->tip_label), tip_text);

		g_free(tip_text);
	}
	else
		gtk_label_set_markup(GTK_LABEL(tooltips->tip_label), data->tip_text);

	gtk_widget_size_request(tooltips->tip_window, &requisition);
	w = requisition.width;
	h = requisition.height;

	gdk_window_get_origin(widget->window, &x, &y);

	if (GTK_WIDGET_NO_WINDOW(widget))
	{
		x += widget->allocation.x;
		y += widget->allocation.y;
	}

	x += widget->allocation.width / 2;

	if (!keyboard_mode)
	{
		gdk_window_get_pointer(gdk_screen_get_root_window(screen),
							   &x, NULL, NULL);
	}

	x -= (w / 2 + 4);

	if ((x + w) > scr_w)
		x -= (x + w) - scr_w;
	else if (x < 0)
		x = 0;

	if ((y + h + widget->allocation.height + 4) > scr_h)
		y = y - h - 4;
	else
		y = y + widget->allocation.height + 4;

	gtk_window_move(GTK_WINDOW(tooltips->tip_window), x, y);
	gtk_widget_show(tooltips->tip_window);
}

static gint
galago_gtk_tooltips_timeout(gpointer data)
{
	GalagoGtkTooltips *tooltips = (GalagoGtkTooltips *)data;

	GDK_THREADS_ENTER();

	if (tooltips->active_tips_data != NULL &&
	    GTK_WIDGET_DRAWABLE(tooltips->active_tips_data->widget))
	{
		galago_gtk_tooltips_draw_tips(tooltips);
	}

	GDK_THREADS_LEAVE();

	return FALSE;
}

static void
galago_gtk_tooltips_set_active_widget(GalagoGtkTooltips *tooltips,
									  GtkWidget *widget)
{
	if (tooltips->tip_window)
	{
		if (GTK_WIDGET_VISIBLE(tooltips->tip_window))
			g_get_current_time(&tooltips->last_popdown);

		gtk_widget_hide(tooltips->tip_window);
	}

	if (tooltips->timer_tag)
	{
		g_source_remove(tooltips->timer_tag);
		tooltips->timer_tag = 0;
	}

	tooltips->active_tips_data = NULL;

	if (widget)
	{
		GList *list;

		for (list = tooltips->tips_data_list; list != NULL; list = list->next)
		{
			GalagoGtkTooltipsData *tooltips_data;

			tooltips_data = list->data;

			if (tooltips_data->widget == widget && GTK_WIDGET_DRAWABLE(widget))
			{
				tooltips->active_tips_data = tooltips_data;
				break;
			}
		}
	}
	else
	{
		tooltips->use_sticky_delay = FALSE;
	}
}

static void
galago_gtk_tooltips_show_tip(GtkWidget *widget)
{
	GalagoGtkTooltipsData *tooltips_data;

	tooltips_data = galago_gtk_tooltips_data_get(widget);

	if (tooltips_data &&
	   (!tooltips_data->tooltips->active_tips_data ||
	     tooltips_data->tooltips->active_tips_data->widget != widget))
	{
		galago_gtk_tooltips_set_active_widget(tooltips_data->tooltips, widget);
		galago_gtk_tooltips_draw_tips(tooltips_data->tooltips);
	}
}

static void
galago_gtk_tooltips_hide_tip(GtkWidget *widget)
{
	GalagoGtkTooltipsData *tooltips_data;

	tooltips_data = galago_gtk_tooltips_data_get(widget);

	if (tooltips_data &&
		(tooltips_data->tooltips->active_tips_data &&
		 tooltips_data->tooltips->active_tips_data->widget == widget))
	{
		galago_gtk_tooltips_set_active_widget(tooltips_data->tooltips,
						  NULL);
	}
}

static gboolean
galago_gtk_tooltips_recently_shown(GalagoGtkTooltips *tooltips)
{
	GTimeVal now;
	glong msec;

	g_get_current_time(&now);

	msec = (now.tv_sec - tooltips->last_popdown.tv_sec) *1000 +
	       (now.tv_usec - tooltips->last_popdown.tv_usec) / 1000;

	return(msec < STICKY_REVERT_DELAY);
}

static gboolean
get_keyboard_mode(GtkWidget *widget)
{
	GtkWidget *toplevel = gtk_widget_get_toplevel(widget);

	if (GTK_IS_WINDOW(toplevel))
	{
		return GPOINTER_TO_UINT(
			g_object_get_data(G_OBJECT(toplevel),
							  "gtk-tooltips-keyboard-mode"));
	}
	else
		return FALSE;
}

static void
start_keyboard_mode(GtkWidget *widget)
{
	GtkWidget *toplevel = gtk_widget_get_toplevel(widget);

	if (GTK_IS_WINDOW(toplevel))
	{
		GtkWidget *focus = GTK_WINDOW(toplevel)->focus_widget;

		g_object_set_data(G_OBJECT(toplevel), "gtk-tooltips-keyboard-mode",
						  GUINT_TO_POINTER(TRUE));

		if (focus)
			galago_gtk_tooltips_show_tip(focus);
	}
}

static void
stop_keyboard_mode(GtkWidget *widget)
{
	GtkWidget *toplevel = gtk_widget_get_toplevel(widget);

	if (GTK_IS_WINDOW(toplevel))
	{
		GtkWidget *focus = GTK_WINDOW(toplevel)->focus_widget;

		if (focus)
			galago_gtk_tooltips_hide_tip(focus);

		g_object_set_data(G_OBJECT(toplevel), "gtk-tooltips-keyboard-mode",
						  GUINT_TO_POINTER(FALSE));
	}
}

static void
galago_gtk_tooltips_start_delay(GalagoGtkTooltips *tooltips, GtkWidget *widget)
{
	GalagoGtkTooltipsData *old_tips_data;

	old_tips_data = tooltips->active_tips_data;

	if (tooltips->enabled &&
	   (!old_tips_data || old_tips_data->widget != widget))
	{
		guint delay;

		galago_gtk_tooltips_set_active_widget(tooltips, widget);

		if (tooltips->use_sticky_delay &&
		    galago_gtk_tooltips_recently_shown(tooltips))
			delay = STICKY_DELAY;
		else
			delay = tooltips->delay;

		tooltips->timer_tag = g_timeout_add(delay,
											galago_gtk_tooltips_timeout,
											(gpointer)tooltips);
	}
}

static void
galago_gtk_tooltips_event_handler(GtkWidget *widget, GdkEvent *event)
{
	GalagoGtkTooltips *tooltips;
	GalagoGtkTooltipsData *old_tips_data;
	GtkWidget *event_widget;
	gboolean keyboard_mode = get_keyboard_mode(widget);

	if ((event->type == GDK_LEAVE_NOTIFY ||
		 event->type == GDK_ENTER_NOTIFY) &&
		event->crossing.detail == GDK_NOTIFY_INFERIOR)
	{
		return;
	}

	old_tips_data = galago_gtk_tooltips_data_get(widget);
	tooltips = old_tips_data->tooltips;

	if (keyboard_mode)
	{
		switch (event->type)
		{
			case GDK_FOCUS_CHANGE:
				if (event->focus_change.in)
					galago_gtk_tooltips_show_tip(widget);
				else
					galago_gtk_tooltips_hide_tip(widget);
				break;

			default:
				break;
		}
	}
	else
	{
		if (event->type != GDK_KEY_PRESS && event->type != GDK_KEY_RELEASE)
		{
			event_widget = gtk_get_event_widget(event);

			if (event_widget != widget)
				return;
		}

		switch (event->type)
		{
			case GDK_EXPOSE:
				/*do nothing */
				break;

			case GDK_ENTER_NOTIFY:
				if (!(GTK_IS_MENU_ITEM(widget) &&
					  GTK_MENU_ITEM(widget)->submenu))
				{
					galago_gtk_tooltips_start_delay(tooltips, widget);
				}

				break;

			case GDK_LEAVE_NOTIFY:
				{
					gboolean use_sticky_delay;

					use_sticky_delay =
						(tooltips->tip_window &&
						 GTK_WIDGET_VISIBLE(tooltips->tip_window));
					galago_gtk_tooltips_set_active_widget(tooltips, NULL);
					tooltips->use_sticky_delay = use_sticky_delay;
				}
				break;
#if 0
			case GDK_MOTION_NOTIFY:

				/*Handle menu items specially ... pend popup for each motion
				 *on other widgets, we ignore motion.
				 */
				if (GTK_IS_MENU_ITEM(widget) &&
					!GTK_MENU_ITEM(widget)->submenu)
				{
					/*Completely evil hack to make sure we get the LEAVE_NOTIFY
					 */
					GTK_PRIVATE_SET_FLAG(widget, GTK_LEAVE_PENDING);
					galago_gtk_tooltips_set_active_widget(tooltips, NULL);
					galago_gtk_tooltips_start_delay(tooltips, widget);
					break;
				}
				break;	/*ignore */
#endif

			case GDK_BUTTON_PRESS:
			case GDK_BUTTON_RELEASE:
			case GDK_KEY_PRESS:
			case GDK_KEY_RELEASE:
			case GDK_PROXIMITY_IN:
			case GDK_SCROLL:
				galago_gtk_tooltips_set_active_widget(tooltips, NULL);
				break;

			default:
				break;
		}
	}
}

static void
galago_gtk_tooltips_widget_unmap(GtkWidget *widget, gpointer data)
{
	GalagoGtkTooltipsData *tooltips_data = (GalagoGtkTooltipsData *)data;
	GalagoGtkTooltips *tooltips = tooltips_data->tooltips;

	if (tooltips->active_tips_data &&
	   (tooltips->active_tips_data->widget == widget))
	{
		galago_gtk_tooltips_set_active_widget(tooltips, NULL);
	}
}

static void
galago_gtk_tooltips_widget_remove(GtkWidget *widget, gpointer data)
{
	GalagoGtkTooltipsData *tooltips_data = (GalagoGtkTooltipsData *)data;
	GalagoGtkTooltips *tooltips = tooltips_data->tooltips;

	galago_gtk_tooltips_widget_unmap(widget, data);

	tooltips->tips_data_list = g_list_remove(tooltips->tips_data_list,
											 tooltips_data);

	galago_gtk_tooltips_destroy_data(tooltips_data);
}

void
_galago_gtk_tooltips_toggle_keyboard_mode(GtkWidget *widget)
{
	if (get_keyboard_mode(widget))
		stop_keyboard_mode(widget);
	else
		start_keyboard_mode(widget);
}
